// SPDX-License-Identifier: MIT
pragma solidity >=0.8.0 <0.9.0;

import "@openzeppelin/contracts-upgradeable/proxy/utils/Initializable.sol";
import "@openzeppelin/contracts-upgradeable/security/PausableUpgradeable.sol";
import "@openzeppelin/contracts-upgradeable/access/AccessControlUpgradeable.sol";
import "@openzeppelin/contracts-upgradeable/proxy/utils/UUPSUpgradeable.sol";
import "@openzeppelin/contracts-upgradeable/security/ReentrancyGuardUpgradeable.sol";
import "@openzeppelin/contracts/utils/cryptography/ECDSA.sol";
import "@openzeppelin/contracts/utils/Context.sol";
import "./AssetVault.sol";
import "./SignatureUtils.sol";
import "./IDapp.sol";
import "./IAdditionalSecurityModule.sol";
import "./libraries/ValsetUpdate.sol";
import "./Utils.sol";
import "hardhat/console.sol";

/**
    @title Facilitates request to Router, validator set update and request from router.
    @author Router Protocol
 */
contract GatewayUpgradeable is
    Initializable,
    PausableUpgradeable,
    AccessControlUpgradeable,
    UUPSUpgradeable,
    ReentrancyGuardUpgradeable
{
    using ValsetUpdate for Utils.ValsetArgs;
    // storage
    string public chainId;
    // address of the vault contract
    AssetVault public vault;
    // current version for encoding decoding of request packet
    uint256 public currentVersion;
    // nonce for requestToRouter Messages
    uint256 public eventNonce;
    uint256 public stateLastValsetNonce;
    // last validated ValSet Checkpoint
    bytes32 public stateLastValsetCheckpoint;
    // default fee for "iSend"
    uint256 public iSendDefaultFee;

    bytes constant INVALID_REQUEST_RESPONSE =  abi.encode(false, abi.encode(""));

    // bytes32 encoding of the string "checkpoint"
    bytes32 constant CHECK_POINT_METHOD_NAME = 0x636865636b706f696e7400000000000000000000000000000000000000000000;

    // bytes32 encoding of "iReceive"
    bytes32 constant IRECEIVE_METHOD_NAME =
    0x6952656365697665000000000000000000000000000000000000000000000000;

    // bytes32 encoding of "iAck"
    bytes32 constant IACK_METHOD_NAME = 0x6941636b00000000000000000000000000000000000000000000000000000000;

    // cross-chain nonce validation mapping for incoming requests from different chains
    // (srcChainid{string} =>  nonce{uint256} => handled{bool}))
    mapping(string => mapping(uint256 => bool)) public nonceExecuted;


    // requestMetadata: 
    // abi.encode (
    //    uint256 destGasLimit;
    //    uint256 destGasPrice;
    //    uint256 ackGasLimit;
    //    uint256 ackGasPrice;
    //    uint256 relayerFees;
    //    Utils.AckType ackType;
    //    bool isReadCall;
    //    bytes asmAddress;
    // )

    // requestPacket:
    // abi.encode (
    //   bytes payload,
    //   bytes destContractAddress
    // )  
    event iSendEvent (
        uint256 version,
        uint256 routeAmount,
        uint256 eventNonce,
        address requestSender,
        string srcChainId,
        string destChainId,
        bytes routeRecipient,
        bytes requestMetadata,
        bytes requestPacket
    );

    event iReceiveEvent (
        uint256 requestIdentifier,
        uint256 indexed eventNonce,
        string srcChainId,
        string destChainId,
        string relayerRouterAddress,
        bytes requestSender,
        bytes contractAckResponses,
        bool status
    );

    // This is CrossTalk Acknowledgement Request Submit Event
    event iAckEvent(
        uint256 indexed eventNonce,
        uint256 indexed requestIdentifier,
        string relayerRouterAddress,
        string chainId,
        bytes data,
        bool success
    );

    event ValsetUpdatedEvent(
        uint256 indexed _newValsetNonce,
        uint256 indexed _eventNonce,
        string srcChainId,
        address[] _validators,
        uint64[] _powers
    );

    event SetDappMetadataEvent(
        uint256 indexed eventNonce,
        bytes dappAddress,
        string chainId,
        string feePayerAddress
    );

    error ExecuteCallsRevert(bool execFlag, bytes execData);

    modifier isSelf() {
        // "ExecuteHandleCalls : Can only be called by Current Contract" => "C01"
        require(_msgSender() == address(this), "C01");
        _;
    }

    function pause() external onlyRole(DEFAULT_ADMIN_ROLE) {
        _pause();
    }

    function unpause() external onlyRole(DEFAULT_ADMIN_ROLE) {
        _unpause();
    }

    // Make a new checkpoint from the supplied validator set
    // A checkpoint is a hash of all relevant information about the valset. This is stored by the contract,
    // instead of storing the information directly. This saves on storage and gas.
    // The format of the checkpoint is:
    // h(_chainId, "checkpoint", valsetNonce, validators[], powers[])
    // Where h is the keccak256 hash function.
    // The validator powers must be decreasing or equal. This is important for checking the signatures on the
    // next valset, since it allows the caller to stop verifying signatures once a quorum of signatures have been verified.
    function makeCheckpoint(Utils.ValsetArgs memory _valsetArgs) internal pure returns (bytes32) {
        bytes32 checkpoint = keccak256(
            abi.encode(CHECK_POINT_METHOD_NAME, _valsetArgs.valsetNonce, _valsetArgs.validators, _valsetArgs.powers)
        );

        return checkpoint;
    }

    // This updates the valset by checking that the validators in the current valset have signed off on the
    // new valset. The signatures supplied are the signatures of the current valset over the checkpoint hash
    // generated from the new valset.
    // Anyone can call this function, but they must supply valid signatures of constantPowerThreshold of the current valset over
    // the new valset.
    function updateValset(
        // The new version of the validator set
        Utils.ValsetArgs calldata _newValset,
        // The current validators that approve the change
        Utils.ValsetArgs calldata _currentValset,
        // These are arrays of the parts of the current validator's signatures
        Utils.Signature[] calldata _sigs
    ) external whenNotPaused nonReentrant {
        // CHECKS
        ValsetUpdate.updateValsetChecks(_newValset, _currentValset);

        // Check that the supplied current validator set matches the saved checkpoint
        if (makeCheckpoint(_currentValset) != stateLastValsetCheckpoint) {
            revert Utils.IncorrectCheckpoint();
        }

        // Check that enough current validators have signed off on the new validator set
        bytes32 newCheckpoint = makeCheckpoint(_newValset);
        bytes32 digest = _make_digest(newCheckpoint);
        SignatureUtils.checkValidatorSignatures(_currentValset, _sigs, digest, Utils.constantPowerThreshold);

        // ACTIONS

        // Stored to be used next time to validate that the valset
        // supplied by the caller is correct.
        stateLastValsetCheckpoint = newCheckpoint;

        // Store new nonce
        stateLastValsetNonce = _newValset.valsetNonce;
        eventNonce++;
        emit ValsetUpdatedEvent(
            _newValset.valsetNonce,
            eventNonce,
            chainId,
            _newValset.validators,
            _newValset.powers
        );
    }

    /**
        @notice Initializes Bridge, creates and grants {msg.sender} the admin role,
        creates and grants {initialRelayers} the relayer role.
        @param _chainId ID of chain the Bridge contract exists on.
     */
    function initialize(
        string memory _chainId,
        address[] memory _validators,
        uint64[] memory _powers,
        uint64 valsetNonce
    ) external initializer {
        // CHECKS
        // Check that validators, powers, and signatures (v,r,s) set is well-formed
        if (_validators.length != _powers.length || _validators.length == 0) {
            revert Utils.MalformedCurrentValidatorSet();
        }

        // Check cumulative power to ensure the contract has sufficient power to actually
        // pass a vote
        // point -> if due to special order our set to
        Utils.ValsetArgs memory _valset;
        _valset = Utils.ValsetArgs(_validators, _powers, valsetNonce);
        ValsetUpdate.validateValsetPower(_valset);

        bytes32 newCheckpoint = makeCheckpoint(_valset);

        // ACTIONS
        __AccessControl_init();
        __Pausable_init();

        // Constructor Fx
        _setupRole(DEFAULT_ADMIN_ROLE, msg.sender);

        chainId = _chainId;
        eventNonce = 0;
        stateLastValsetNonce = valsetNonce;
        stateLastValsetCheckpoint = newCheckpoint;
        eventNonce++;
        currentVersion = 1;
        emit ValsetUpdatedEvent(valsetNonce, eventNonce, chainId, _validators, _powers);
        // Constructor Fx
    }

    function _authorizeUpgrade(address newImplementation) internal virtual override onlyRole(DEFAULT_ADMIN_ROLE) {}

    function setCurrentVersion(uint256 newVersion) external onlyRole(DEFAULT_ADMIN_ROLE) {
        currentVersion = newVersion;
    }

    function setBridgeFees(
        uint256 _iSendDefaultFee
    ) external onlyRole(DEFAULT_ADMIN_ROLE) {
        // default fee for "sendCrossChainRequest"
        iSendDefaultFee = _iSendDefaultFee;
    }

    function setVault(address _vaultAddress) external onlyRole(DEFAULT_ADMIN_ROLE) {
        // "empty address" => "C02"
        require(_vaultAddress != address(0), "C02");
        vault = AssetVault(_vaultAddress);
    }

    function setDappMetadata(string memory feePayerAddress) external payable nonReentrant returns (uint256) {
        // "fees too low" => "C03"
        require(msg.value >= iSendDefaultFee, "C03");
        eventNonce++;
        emit SetDappMetadataEvent(eventNonce, toBytes(msg.sender), chainId, feePayerAddress);
        return eventNonce;
    }

    // TODO: Memory or calldata check gas
    function iSend(
        uint256 version,
        Utils.NativeTransferParams memory nativeTransferParams,
        string memory destChainId,
        bytes calldata requestMetadata,
        bytes calldata requestPacket
    ) external payable whenNotPaused returns (uint256) {
        // "fees too low" => "C03"
        require(msg.value >= iSendDefaultFee, "C03");

        if (nativeTransferParams.routeAmount > 0) {
            // "empty recipient" => "C04"
            // keccak256("") = 0xc5d2460186f7233c927e7db2dcc703c0e500b653ca82273b7bfad8045d85a470
            // TODO: Can we remove
            require(keccak256(nativeTransferParams.routeRecipient) != 0xc5d2460186f7233c927e7db2dcc703c0e500b653ca82273b7bfad8045d85a470, 
                "C04"
            );

            vault.deposit(nativeTransferParams.routeAmount, msg.sender);
        }

        eventNonce++;

        emit iSendEvent(
            version,
            nativeTransferParams.routeAmount, 
            eventNonce, 
            msg.sender, 
            chainId,
            destChainId, 
            nativeTransferParams.routeRecipient,
            requestMetadata,
            requestPacket
        );

        return eventNonce;
    }

    function checkCheckpointAndVerifySigs(
        Utils.ValsetArgs calldata _valsetArgs,
        Utils.Signature[] calldata _sigs,
        bytes memory encodedABI
    ) internal view {
        if (makeCheckpoint(_valsetArgs) != stateLastValsetCheckpoint) {
            revert Utils.IncorrectCheckpoint();
        }

        bytes32 messagehash = keccak256(encodedABI);
        bytes32 digest = _make_digest(messagehash);
        SignatureUtils.checkValidatorSignatures(_valsetArgs, _sigs, digest, Utils.constantPowerThreshold);
    }

    function _make_digest(bytes32 data) private pure returns (bytes32 _digest) {
        _digest = keccak256(abi.encodePacked(Utils.MSG_PREFIX, data));
    }

    // TODO: Remove gasleft
    function iReceive(
        Utils.ValsetArgs calldata _currentValset,
        Utils.Signature[] calldata _sigs,			    
        Utils.RequestPayload memory requestPayload,
        string memory relayerRouterAddress
    ) external whenNotPaused nonReentrant {
        uint256 gasLeft = gasleft();

        // TODO: Change individual fields to the whole payload 
        bytes memory encodedABI = abi.encode(
            IRECEIVE_METHOD_NAME,
            requestPayload.routeAmount,
            requestPayload.requestIdentifier,
            requestPayload.requestTimestamp,
            requestPayload.routeRecipient,
            requestPayload.asmAddress,
            requestPayload.srcChainId,
            requestPayload.destChainId,
            requestPayload.requestSender,
            requestPayload.requestPacket,
            requestPayload.isReadCall
        );

        checkCheckpointAndVerifySigs(_currentValset, _sigs, encodedABI);

        require(keccak256(bytes(chainId)) == keccak256(bytes(requestPayload.destChainId)), "wrong dest ID");

        // "cross-chain request message already handled" => "C06"
        require(nonceExecuted[requestPayload.srcChainId][requestPayload.requestIdentifier] == false, "C06");

        if (requestPayload.routeAmount > 0) {
            //caution: make sure recipient should be correct
            require(requestPayload.routeRecipient != address(0), "Recipient = addr(0)");
            vault.handleWithdraw(requestPayload.routeAmount, requestPayload.routeRecipient);
        }

        (bytes memory packet, bytes memory handler) = abi.decode(
            requestPayload.requestPacket,
            (bytes, bytes)
        );

        bool isValidRequest = true;

        if (requestPayload.asmAddress != address(0) && isContract(requestPayload.asmAddress)) {
            IAdditionalSecurityModule asm = IAdditionalSecurityModule(requestPayload.asmAddress);
            isValidRequest = asm.verifyCrossChainRequest(
                requestPayload.requestIdentifier,
                requestPayload.requestTimestamp,
                requestPayload.requestSender,
                requestPayload.srcChainId,
                packet,
                handler
            );
        }

        nonceExecuted[requestPayload.srcChainId][requestPayload.requestIdentifier] = true;
        eventNonce++;


        if (!isValidRequest) {
            // Since we are sending a specific set of bytes (invalid request response), we can remove exeCode
            emit iReceiveEvent(
                requestPayload.requestIdentifier,
                eventNonce,
                requestPayload.srcChainId,
                chainId,
                relayerRouterAddress,
                requestPayload.requestSender,
                INVALID_REQUEST_RESPONSE,
                false
            );
            return;
        }

        (bool success, bytes memory data) = address(this).call(
            abi.encodeWithSelector(
                this.executeHandlerCalls.selector,
                requestPayload.requestSender,
                requestPayload.srcChainId,
                handler,
                packet,
                requestPayload.isReadCall
            )
        );

        emit iReceiveEvent(
            requestPayload.requestIdentifier,
            eventNonce,
            requestPayload.srcChainId,
            chainId,
            relayerRouterAddress,
            requestPayload.requestSender,
            data,
            success
        );

        console.log("gas left handle request: %s", gasLeft - gasleft());
    }

    function executeHandlerCalls(
        string memory requestSender,
        string memory srcChainId,
        bytes memory handler,
        bytes memory payload,
        bool isReadCall
    ) external isSelf returns (bool, bytes memory) {
        bytes memory execData;
        address handlerAddr = toAddress(handler);
        bool execFlag = isContract(handlerAddr);

        if (execFlag) {
            if (!isReadCall) {
                (execFlag, execData) = handlerAddr.call(
                    abi.encodeWithSelector(
                        IDapp.iReceive.selector,
                        requestSender,
                        payload,
                        srcChainId
                    )
                );
            } else {
                (execFlag, execData) = handlerAddr.call(payload);
            }
        }

        if (!execFlag) {
            revert ExecuteCallsRevert({ execFlag: execFlag, execData: execData });
        }

        return (execFlag, execData);
    }

    // TODO: gas left remove
    function iAck(
        // The validators that approve the call
        Utils.ValsetArgs calldata _currentValset,
        // These are arrays of the parts of the validators signatures
        Utils.Signature[] calldata _sigs,
        Utils.CrossChainAckPayload memory crossChainAckPayload
    ) external whenNotPaused nonReentrant {
        uint256 gasLeft = gasleft();
        // chainId, chainType data is present inside the signature validation
        // valsetArgs validation will be handled by the makeCheckPoint function
        require(
            nonceExecuted[crossChainAckPayload.destChainId][
                crossChainAckPayload.requestIdentifier
            ] == false,
            "C06"
        );

        // TODO: Change individual fields to the whole payload 
        bytes memory encodedABI = abi.encode(
            // bytes32 encoding of "iAck"
            IACK_METHOD_NAME,
            chainId,
            crossChainAckPayload.requestIdentifier, 
            crossChainAckPayload.relayerRouterAddress,
            crossChainAckPayload.destChainId,
            crossChainAckPayload.requestSender,
            crossChainAckPayload.execData,
            crossChainAckPayload.execFlag
        );

        checkCheckpointAndVerifySigs(_currentValset, _sigs, encodedABI);

        nonceExecuted[crossChainAckPayload.destChainId][
            crossChainAckPayload.requestIdentifier
        ] = true;
        eventNonce++;

        (bool success, bytes memory data) = address(this).call(
            abi.encodeWithSelector(
                this.executeCrossTalkAck.selector,
                crossChainAckPayload.requestIdentifier,
                crossChainAckPayload.requestSender,
                crossChainAckPayload.execFlag,
                crossChainAckPayload.execData
            )
        );

        emit iAckEvent(
            eventNonce,
            crossChainAckPayload.requestIdentifier,
            crossChainAckPayload.relayerRouterAddress,
            chainId,
            data,
            success
        );

        console.log("gas left ack: %s", gasLeft - gasleft());
    }

    function executeCrossTalkAck(
        uint64 requestIdentifier,
        bytes memory requestSender,
        bool execFlag,
        bytes memory execData
    ) external isSelf {
        address handler = toAddress(requestSender);
        bool success;
        bytes memory data;
        success = isContract(handler);
        if (success) {
            (success, data) = handler.call(
                abi.encodeWithSelector(
                    IDapp.iAck.selector,
                    requestIdentifier,
                    execFlag,
                    execData
                )
            );
        }
    }

    function isContract(address addr) internal view returns (bool) {
        uint256 size;
        assembly {
            size := extcodesize(addr)
        }
        return size > 0;
    }

    function toAddress(bytes memory _bytes) internal pure returns (address contractAddress) {
        bytes20 srcTokenAddress;
        assembly {
            srcTokenAddress := mload(add(_bytes, 0x20))
        }
        contractAddress = address(srcTokenAddress);
    }

    function toBytes(address a) public pure returns (bytes memory b) {
        assembly {
            let m := mload(0x40)
            a := and(a, 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF)
            mstore(add(m, 20), xor(0x140000000000000000000000000000000000000000, a))
            mstore(0x40, add(m, 52))
            b := m
        }
    }

    receive() external payable {}
}
